/*
 * Nenad Taskovic 5/93
 * e-mail: neyt@fon.bg.ac.yu
 * www: http://fon.fon.bg.ac.yu/~neyt/
 */

/*********************************************************************/
/** deklaracija funkcija i podataka za rad sa bilo kojom listom */
typedef struct Cvor {
	void *podatak;
   struct Cvor *pret, *sled;
} Cvor;

typedef struct Lista {
	Cvor *glava, *tek;
   int duzpod;
} Lista;

Lista *kreiraj_listu (int duzpod);
void unisti_listu (Lista *lista);
void ubaci_na_pocetak (Lista *lista, void *podatak);
void ubaci_na_kraj (Lista *lista, void *podatak);
void *podatak (Lista *lista);
void na_pocetak (Lista *lista);
void napred (Lista *lista);
void na_kraj (Lista *lista);
void nazad (Lista *lista);
int kraj_liste (Lista *lista);
void izbaci_iz_liste (Lista *lista);
void izbaci_sa_pocetka (Lista *lista);
void izbaci_sa_kraja (Lista *lista);

/*********************************************************************/
/** deklaracija funkcija za rad sa grafikom */

typedef struct {
	int x, y;
} Tacka;

void nacrtaj_koord_sistem (void);
void nacrtaj_grafik (Tacka *tacka, int tacaka);
void zatvori_grafiku (void);

/*********************************************************************/
/** implementacija */

#include <values.h>
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <graphics.h>


Lista *kreiraj_listu (int duzpod)
{
	Lista *lista;
	Cvor *dummy;

	lista = (Lista *) malloc(sizeof(Lista));
	lista->duzpod = duzpod;
	dummy = (Cvor *) malloc(sizeof(Cvor));
	dummy->pret = dummy;
	dummy->sled = dummy;
	dummy->podatak = 0;
	lista->glava = dummy;

	return lista;
}

Cvor *alociraj_cvor (Lista *lista, void *podatak)
{
   Cvor *cvor;

   cvor = (Cvor *) malloc(sizeof(Cvor));
	cvor->podatak = malloc(lista->duzpod);
   memcpy(cvor->podatak, podatak, lista->duzpod);

   return cvor;
}

void dealociraj_cvor (Cvor *cvor)
{
   free(cvor->podatak);
   free(cvor);
}

void unisti_listu (Lista *lista)
{
   Cvor *cvor, *tmp;
   
   cvor = lista->glava->sled;
   lista->glava->sled = NULL;
   while (cvor) {
      tmp = cvor;
		cvor = cvor->sled;
      dealociraj_cvor(tmp);
   }
   free(lista);
}

void ubaci_na_pocetak (Lista *lista, void *podatak)
{
   Cvor *cvor;
   
   cvor = alociraj_cvor(lista, podatak);
   cvor->pret = lista->glava;
	cvor->sled = lista->glava->sled;
   cvor->sled->pret = cvor;
   lista->glava->sled = cvor;
}

void ubaci_na_kraj (Lista *lista, void *podatak)
{
   Cvor *cvor;

   cvor = alociraj_cvor(lista, podatak);
   cvor->sled = lista->glava;
   cvor->pret = lista->glava->pret;
   cvor->pret->sled = cvor;
   lista->glava->pret = cvor;
}

void *podatak (Lista *lista)
{
   return lista->tek->podatak;
}

void na_pocetak (Lista *lista)
{
   lista->tek = lista->glava->sled;
}

void napred (Lista *lista)
{
	lista->tek = lista->tek->sled;
}

void na_kraj (Lista *lista)
{
   lista->tek = lista->glava->pret;
}

void nazad (Lista *lista)
{
   lista->tek = lista->tek->pret;
}

int kraj_liste (Lista *lista)
{
   return (lista->tek == lista->glava);
}

void izbaci_iz_liste (Lista *lista)
{
	lista->tek->pret->sled = lista->tek->sled;
   lista->tek->sled->pret = lista->tek->pret;
   dealociraj_cvor(lista->tek);
}

void izbaci_sa_pocetka (Lista *lista)
{
   Cvor *cvor;
   
   cvor = lista->glava->sled;
   lista->glava->sled = cvor->sled;
   cvor->sled->pret = lista->glava;
	dealociraj_cvor(cvor);
}

void izbaci_sa_kraja (Lista *lista)
{
   Cvor *cvor;
   
   cvor = lista->glava->pret;
	lista->glava->pret = cvor->pret;
   cvor->pret->sled = lista->glava;
	dealociraj_cvor(cvor);
}

/*********************************************************************/
/** rad sa grafikom */

#define MARGINA 20
#define VC 3
#define RAZMAK 20

void nacrtaj_koord_sistem (void)
{
	int maxx, maxy, i;
	char broj[10];
	int gdriver = DETECT, gmode, errorcode;
	initgraph(&gdriver, &gmode, "");
	errorcode = graphresult();
	if (errorcode != grOk) {
		printf("Greska: %s\n", grapherrormsg(errorcode));
		exit(1);
	}
	maxx = getmaxx();
	maxy = getmaxy();

	line(MARGINA, 0, MARGINA, maxy);
	line(0, maxy-MARGINA, maxx, maxy-MARGINA);
	for (i=0; i<=32; i+=2) {
		line(MARGINA+i*18, maxy-MARGINA-VC, MARGINA+i*18, maxy-MARGINA+VC);
		outtextxy(MARGINA+i*18-2*VC, maxy-MARGINA+VC*3, itoa(i*2, broj, 10));
	}
}

void nacrtaj_grafik(Tacka *tacka, int tacaka)
{
	int maxy=getmaxy();
	int i, x, y;

	moveto(MARGINA+2*VC, MARGINA+2*VC);
	for (i=0; i<tacaka; i++) {
		x = tacka[i].x*9;
		y = tacka[i].y/4;
		lineto(MARGINA+x, maxy-MARGINA-y);
	}
}

void zatvori_grafiku (void)
{
	closegraph();
}

/**************************************************************************/
/** deo koda sa algoritmima za simulaciju upravljanja memorijom */

typedef struct {
	int stranica, od_kraja;
} Frame;


int nije_u_listi(Lista *lista, int strana)
{
	na_pocetak(lista);
	while ( !kraj_liste(lista) && (strana != *((int *)podatak(lista))) )
		napred(lista);

	return kraj_liste(lista);
}

int FIFO (int *trag, int duzina_traga, int broj_okvira)
{
	Lista *lista;
	int i, broj = -1 , prekida = 0;

	lista = kreiraj_listu(sizeof(int));
	for (i=0; i<broj_okvira; i++)			/** punjenje liste praznim okvirima */
		ubaci_na_pocetak(lista, &broj);	/** strana */
	for (i=0; i<duzina_traga; i++) {
	 if (nije_u_listi(lista, trag[i])) { /** ako stranica nije u listi okvira */
			izbaci_sa_kraja(lista);			 /** izbacuje se stranica sa kraja liste */
			ubaci_na_pocetak(lista, &(trag[i])); /** a na pocetak ubacuje nova */
			prekida++;
		}
	}
	unisti_listu(lista);

	return prekida;
}

int LRU (int *trag, int duzina_traga, int broj_okvira)
{
	Lista *lista;
	int i, broj = -1 , prekida = 0;

	lista = kreiraj_listu(sizeof(int));
	for (i=0; i<broj_okvira; i++)			/** punjenje liste praznim okvirima */
		ubaci_na_pocetak(lista, &broj);	/** strana */
	for (i=0; i<duzina_traga; i++) {
		if (nije_u_listi(lista, trag[i])) { /** ako stranica nije u listi okvira */
			izbaci_sa_kraja(lista);				/** izbacuje se stranica sa kraja liste */
			prekida++;
		} else
			izbaci_iz_liste(lista);		/** ako jeste izbacuje se sa mesta na kojem */
		ubaci_na_pocetak(lista, &(trag[i])); /** je pronadjena i ubacuje ponovo na pocetak */
	}
	unisti_listu(lista);

	return prekida;
}

int nije_u_listi_OPT (Lista *lista, int strana)
{
	na_pocetak(lista);
	while ( !kraj_liste(lista) && (strana != ((Frame *)podatak(lista))->stranica) )
		napred(lista);

	return kraj_liste(lista);
}

void pronadji_najgoru_i_zameni_podatke (Lista *lista, Frame *frame)
{
	int min = MAXINT;
	Frame *minf, *tek;

	na_pocetak(lista);
	while ( !kraj_liste(lista)) {		/** pronalazi se najmanja udaljenost */
		tek = (Frame *) podatak(lista); /** prve pojave straniice od kraja traga */
		if (tek->od_kraja < min) {		/** tj. stranica u memoriji koja ce se */
			min = tek->od_kraja;			/** najkasnije ponovo javiti */
			minf = tek;
		}
		napred(lista);
	}
	*minf = *frame;
}

int OPT (int *trag, int duzina_traga, int broj_okvira)
{
   Lista *lista;
   Frame frame, *frmpok;
   int i, prekida = 0, j;

   lista = kreiraj_listu(sizeof(Frame));

   frame.stranica = -1;
   frame.od_kraja = -1;
	for (i=0; i<broj_okvira; i++)			/** punjenje liste praznim okvirima */
		ubaci_na_pocetak(lista, &frame); /** strana */
	for (i=0; i<duzina_traga; i++) {	/** za svaku stranicu koja se ubacuje u memoriju */
			for (j=i+1; (j<duzina_traga) && (trag[j] != trag[i]); j++);
			j = duzina_traga - j;	/** pronalazi se njena sledeca pojava u tragu, */
			frame.stranica = trag[i]; /** tj. udaljenost sledece pojave od kraja */
			frame.od_kraja = j;		  /** traga */
		if (nije_u_listi_OPT(lista, trag[i])) { /** ako stranica nije u listi (memoriji) */
			pronadji_najgoru_i_zameni_podatke(lista, &frame);
			prekida++; /** pronalazi se najgora stranica i u njen okvir se smesta */
		} else {		/** nova stranica */
			frmpok = (Frame *)podatak(lista);
			*frmpok = frame;	/** ako se vec nalazi u listi azuriraju se podaci o */
		}							/** sledecoj pojavi te stranice u tragu */
	}
	unisti_listu(lista);

	return prekida;
}

/*********************************************************************/
/** prikaz podataka */

void ispisi_tabelu (int p[3][4], int dt)
{
	puts("       FIFO             LRU              OPT");
	puts("Ŀ");
 printf("  8  %5d  %5.2f%%  %5d  %5.2f%%  %5d  %5.2f%% \n",
	 p[0][0], p[0][0]*100.0/dt, p[1][0], p[1][0]*100.0/dt, p[2][0], p[2][0]*100.0/dt);
 printf(" 16  %5d  %5.2f%%  %5d  %5.2f%%  %5d  %5.2f%% \n",
	 p[0][1], p[0][1]*100.0/dt, p[1][1], p[1][1]*100.0/dt, p[2][1], p[2][1]*100.0/dt);
 printf(" 32  %5d  %5.2f%%  %5d  %5.2f%%  %5d  %5.2f%% \n",
	 p[0][2], p[0][2]*100.0/dt, p[1][2], p[1][2]*100.0/dt, p[2][2], p[2][2]*100.0/dt);
 printf(" 64  %5d  %5.2f%%  %5d  %5.2f%%  %5d  %5.2f%% \n",
	 p[0][3], p[0][3]*100.0/dt, p[1][3], p[1][3]*100.0/dt, p[2][3], p[2][3]*100.0/dt);
	puts("");
}

int main (int argc, char *argv[])
{
	FILE *dat;
	int i, trag[2000], okvira;
	int prekida[3][4];
	Tacka tacka[64];

   if (argc < 2) {
      printf("Nacin koriscenja:\n%s <datoteka sa tragom>\n", argv[0]);
      return 1;
   }

   dat = fopen(argv[1], "rt");
   if (dat == NULL) {
      printf("Datoteka sa imenom `%s' ne postoji.\n", argv[1]);
      return 3;
   }

   for (i=0; i<2000; i++)
		fscanf(dat, " %d", &(trag[i]));

   fclose(dat);

	okvira = 8;
	for (i=0; i<4; i++) {
      prekida[0][i] = FIFO(trag, 2000, okvira);
      okvira *= 2;
   }
   okvira = 8;
   for (i=0; i<4; i++) {
      prekida[1][i] = LRU(trag, 2000, okvira);
      okvira *= 2;
   }
   okvira = 8;
   for (i=0; i<4; i++) {
      prekida[2][i] = OPT(trag, 2000, okvira);
      okvira *= 2;
   }
   ispisi_tabelu(prekida, 2000);
   printf("\n");
   getch();
	nacrtaj_koord_sistem();
	setcolor(1);
	for (i=1; i<65; i++) {
		tacka[i-1].x = i;
		tacka[i-1].y = FIFO(trag, 2000, i);
	}
	nacrtaj_grafik(tacka, 64);
   setcolor(2);
	for (i=1; i<65; i++) {
		tacka[i-1].x = i;
		tacka[i-1].y = LRU(trag, 2000, i);
	}
	nacrtaj_grafik(tacka, 64);
	setcolor(3);
	for (i=1; i<65; i++) {
		tacka[i-1].x = i;
		tacka[i-1].y = OPT(trag, 2000, i);
	}
	nacrtaj_grafik(tacka, 64);
	getch();
   zatvori_grafiku();

	return 0;
}

